import warnings

import cv2
import numpy as np

from .winctrl import WinCtrl
from ..webcam_pub.camctrl import CamCtrl
from localpubsub import NoData

if False:
    from typing import List, Union, Callable, Any
    import numpy as np


class SubscriberWindows(object):
    frame_dict = {}

    esc_key_codes = [27]  # ESC key on most keyboards

    def __init__(self,
                 window_names=('cvpubsubs',),  # type: List[str]
                 video_sources=(0,),  # type: List[Union[str,int]]
                 callbacks=(None,),  # type: List[Callable[[List[np.ndarray]], Any]]
                 ):
        self.window_names = window_names
        self.source_names = []
        for name in video_sources:
            if isinstance(name, np.ndarray):
                self.source_names.append(str(hash(str(name))))
                self.input_vid_global_names = [str(hash(str(name))) + "frame" for name in video_sources]
            elif len(str(name))<=1000:
                self.source_names.append(str(name))
                self.input_vid_global_names = [str(name) + "frame" for name in video_sources]
            else:
                raise ValueError("Input window name too long.")

        self.callbacks = callbacks
        self.input_cams = video_sources


    @staticmethod
    def set_global_frame_dict(name, *args):
        if len(str(name)) <= 1000:
            SubscriberWindows.frame_dict[str(name) + "frame"] = [*args]
        elif isinstance(name, np.ndarray):
            SubscriberWindows.frame_dict[str(hash(str(name))) + "frame"] = [*args]
        else:
            raise ValueError("Input window name too long.")


    def __stop_all_cams(self):
        for c in self.source_names:
            CamCtrl.stop_cam(c)

    def handle_keys(self,
                    key_input,  # type: int
                    ):
        if key_input in self.esc_key_codes:
            for name in self.window_names:
                cv2.destroyWindow(name + " (press ESC to quit)")
            WinCtrl.quit()
            self.__stop_all_cams()
            return 'quit'
        elif key_input not in [-1, 0]:
            try:
                WinCtrl.key_pub.publish(chr(key_input))
            except ValueError:
                warnings.warn(
                    RuntimeWarning("Unknown key code: [{}]. Please report to cv_pubsubs issue page.".format(key_input))
                )

    def update_window_frames(self):
        win_num = 0
        for i in range(len(self.input_vid_global_names)):
            if self.input_vid_global_names[i] in self.frame_dict and not isinstance(self.frame_dict[
                                                                        self.input_vid_global_names[i]], NoData):
                if len(self.callbacks)>0 and self.callbacks[i % len(self.callbacks)] is not None:
                    frames = self.callbacks[i % len(self.callbacks)](self.frame_dict[self.input_vid_global_names[i]])
                else:
                    frames = self.frame_dict[self.input_vid_global_names[i]]
                if isinstance(frames, np.ndarray) and len(frames.shape)<=3:
                    frames = [frames]
                for f in range(len(frames)):
                    cv2.imshow(self.window_names[win_num % len(self.window_names)] + " (press ESC to quit)", frames[f])
                    win_num += 1

    # todo: figure out how to get the red x button to work. Try: https://stackoverflow.com/a/37881722/782170
    def loop(self):
        sub_cmd = WinCtrl.win_cmd_sub()
        msg_cmd = ''
        key = ''
        while msg_cmd != 'quit' and key != 'quit':
            self.update_window_frames()
            msg_cmd = sub_cmd.get()
            key = self.handle_keys(cv2.waitKey(1))
        sub_cmd.release()
        WinCtrl.quit(force_all_read=False)
        self.__stop_all_cams()



import threading
import time


import cv2
import numpy as np

from cvpubsubs.webcam_pub.camctrl import CamCtrl
from .np_cam import NpCam

if False:
    from typing import Union, Tuple


def pub_cam_loop(cam_id,  # type: Union[int, str]
                 request_size=(1280, 720),  # type: Tuple[int, int]
                 high_speed=False,  # type: bool
                 fps_limit=240  # type: float
                 ):  # type: (...)->bool
    """Publishes whichever camera you select to CVCams.<cam_id>.Vid
    You can send a quit command 'quit' to CVCams.<cam_id>.Cmd
    Status information, such as failure to open, will be posted to CVCams.<cam_id>.Status


    :param high_speed: Selects mjpeg transferring, which most cameras seem to support, so speed isn't limited
    :param fps_limit: Limits the frames per second.
    :param cam_id: An integer representing which webcam to use, or a string representing a video file.
    :param request_size: A tuple with width, then height, to request the video size.
    :return: True if loop ended normally, False if it failed somehow.
    """

    if isinstance(cam_id, (int, str)):
        name = str(cam_id)
    elif isinstance(cam_id, np.ndarray):
        name = str(hash(str(cam_id)))
    else:
        raise TypeError("Only strings or ints representing cameras, or numpy arrays representing pictures supported.")

    if isinstance(cam_id, (int, str)):
        cam = cv2.VideoCapture(cam_id)
    elif isinstance(cam_id, np.ndarray):
        cam = NpCam(cam_id)  # type: NpCam
    else:
        raise TypeError("Only strings or ints representing cameras, or numpy arrays representing pictures supported.")

    CamCtrl.register_cam(name)

    # cam.set(cv2.CAP_PROP_CONVERT_RGB, 0)
    frame_counter = 0

    sub = CamCtrl.cam_cmd_sub(name)
    sub.return_on_no_data = ''
    msg = ''

    if high_speed:
        cam.set(cv2.CAP_PROP_FOURCC, cv2.CAP_OPENCV_MJPEG)

    cam.set(cv2.CAP_PROP_FRAME_WIDTH, request_size[0])
    cam.set(cv2.CAP_PROP_FRAME_HEIGHT, request_size[1])

    if not cam.isOpened():
        CamCtrl.cv_cams_dict[name].status_pub.publish("failed")
        return False
    now = time.time()
    while msg != 'quit':
        time.sleep(1. / (fps_limit - (time.time() - now)))
        now = time.time()
        (ret, frame) = cam.read()  # type: Tuple[bool, np.ndarray ]
        if ret is False or not isinstance(frame, np.ndarray):
            cam.release()
            CamCtrl.cv_cams_dict[name].status_pub.publish("failed")
            return False
        if cam.get(cv2.CAP_PROP_FRAME_COUNT) > 0:
            frame_counter += 1
            if frame_counter >= cam.get(cv2.CAP_PROP_FRAME_COUNT):
                frame_counter = 0
                cam = cv2.VideoCapture(cam_id)
        CamCtrl.cv_cams_dict[name].frame_pub.publish(frame)
        msg = sub.get()
    sub.release()

    cam.release()
    return True


def pub_cam_thread(cam_id,  # type: Union[int, str]
                   request_ize=(1280, 720),  # type: Tuple[int, int]
                   high_speed=False,  # type: bool
                   fps_limit=240  # type: float
                   ):
    # type: (...) -> threading.Thread
    t = threading.Thread(target=pub_cam_loop, args=(cam_id, request_ize, high_speed, fps_limit))
    t.start()
    return t



import numpy as np
import cv2


class NpCam(object):
    def __init__(self, img):
        assert isinstance(img, np.ndarray)
        self.__img = img
        self.__is_opened = True

        self.__width = self.__img.shape[1]
        self.__height = self.__img.shape[0]
        self.__ratio = float(self.__width) / self.__height

        self.__wait_for_ratio = False

    def set(self, *args, **kwargs):
        if args[0] in [cv2.CAP_PROP_FRAME_WIDTH, cv2.CAP_PROP_FRAME_HEIGHT]:
            self.__wait_for_ratio = not self.__wait_for_ratio
            if args[0] == cv2.CAP_PROP_FRAME_WIDTH:
                self.__width = args[1]
            else:
                self.__height = args[1]
            if not self.__wait_for_ratio:
                if self.__width <= 0 or not isinstance(self.__width, int):
                    self.__width = int(self.__ratio * self.__height)
                elif self.__height <= 0 or not isinstance(self.__height, int):
                    self.__height = int(self.__width / self.__ratio)
                if self.__width>0 and self.__height>0:
                    self.__img = cv2.resize(self.__img, (self.__width, self.__height))

    def get(self, *args, **kwargs):
        if args[0] == cv2.CAP_PROP_FRAME_COUNT:
            return float("inf")

    def read(self):
        return (True, self.__img)

    def isOpened(self):
        return self.__is_opened

    def release(self):
        self.__is_opened = False



import threading
import logging

from localpubsub import VariablePub, VariableSub


class WinCtrl(object):
    key_pub = VariablePub()
    win_cmd_pub = VariablePub()

    @staticmethod
    def quit(force_all_read=True):
        WinCtrl.win_cmd_pub.publish('quit', force_all_read=force_all_read)

    @staticmethod
    def win_cmd_sub(): # type: ()->VariableSub
        return WinCtrl.win_cmd_pub.make_sub()  # type: VariableSub



import threading
import unittest as ut

import numpy as np

import cvpubsubs.webcam_pub as w
from cvpubsubs.window_sub import SubscriberWindows
from cvpubsubs.window_sub.winctrl import WinCtrl

if False:
    import numpy as np


def print_keys_thread():
    sub_key = WinCtrl.key_pub.make_sub()
    sub_cmd = WinCtrl.win_cmd_pub.make_sub()
    sub_cmd.return_on_no_data = ''
    msg_cmd = ''
    while msg_cmd != 'quit':
        key_chr = sub_key.get(sub_key)  # type: np.ndarray
        WinCtrl.key_pub.publish(None) # consume data
        if key_chr is not None:
            print("key pressed: " + str(key_chr))
        msg_cmd = sub_cmd.get()
    WinCtrl.quit(force_all_read=False)


def start_print_keys_thread():  # type: (...) -> threading.Thread
    t = threading.Thread(target=print_keys_thread, args=())
    t.start()
    return t


class TestSubWin(ut.TestCase):

    def test_sub(self):
        w.VideoHandlerThread().display()

    def test_image(self):
        img = np.random.uniform(0, 1, (300, 300, 3))
        w.VideoHandlerThread(video_source=img).display()

    def test_sub_with_args(self):
        video_thread = w.VideoHandlerThread(video_source=0,
                                            callbacks=w.display_callbacks,
                                            request_size=(800, 600),
                                            high_speed=False,
                                            fps_limit=8
                                            )

        video_thread.display()

    def test_sub_with_callback(self):
        def redden_frame_print_spam(frame, cam_id):
            frame[:, :, 0] = 0
            frame[:, :, 2] = 0

        w.VideoHandlerThread(callbacks=[redden_frame_print_spam] + w.display_callbacks).display()

    def test_multi_cams_one_source(self):
        def cam_handler(frame, cam_id):
            SubscriberWindows.set_global_frame_dict(cam_id, frame, frame)

        t = w.VideoHandlerThread(0, [cam_handler],
                                 request_size=(1280, 720),
                                 high_speed=True,
                                 fps_limit=240
                                 )

        t.start()

        SubscriberWindows(window_names=['cammy', 'cammy2'],
                          video_sources=[str(0)]
                          ).loop()

        t.join()

    @ut.skip("I don't have stereo cams... :(")
    def test_multi_cams_multi_source(self):
        t1 = w.VideoHandlerThread(0, request_size=(1920, 1080))
        t2 = w.VideoHandlerThread(1, request_size=(1920, 1080))

        t1.start()
        t2.start()

        SubscriberWindows(window_names=['cammy', 'cammy2'],
                          video_sources=[0, 1]
                          ).loop()

        t1.join()
        t1.join()



from threading import Lock
from localpubsub import VariablePub, VariableSub

if False:
    from typing import Union, Dict


class CamHandler(object):
    def __init__(self, name, sub):
        self.name = name
        self.cmd = None
        self.sub = sub  # type: VariableSub
        self.pub = VariablePub()
        self.cmd_pub = VariablePub()


class Cam(object):
    def __init__(self, name):
        self.name = name
        self.cmd = None
        self.frame_pub = VariablePub()
        self.cmd_pub = VariablePub()
        self.status_pub = VariablePub()


class CamCtrl(object):
    cv_cam_handlers_dict = {}  # type: Dict[str, CamHandler]
    cv_cams_dict = {}  # type: Dict[str, Cam]

    @staticmethod
    def register_cam(cam_id):
        cam = Cam(str(cam_id))
        CamCtrl.cv_cams_dict[str(cam_id)] = cam
        CamCtrl.cv_cam_handlers_dict[str(cam_id)] = CamHandler(str(cam_id), cam.frame_pub.make_sub())

    @staticmethod
    def stop_cam(cam_id  # type: Union[int, str]
                 ):
        CamCtrl.cv_cams_dict[str(cam_id)].cmd_pub.publish('quit', blocking=True)
        CamCtrl.cv_cam_handlers_dict[str(cam_id)].cmd_pub.publish('quit', blocking=True)

    @staticmethod
    def cam_cmd_sub(cam_id, blocking=True):
        if blocking:
            while cam_id not in CamCtrl.cv_cams_dict:
                continue
        return CamCtrl.cv_cams_dict[str(cam_id)].cmd_pub.make_sub()

    @staticmethod
    def cam_frame_sub(cam_id, blocking=True):
        if blocking:
            while cam_id not in CamCtrl.cv_cams_dict:
                continue
        return CamCtrl.cv_cams_dict[str(cam_id)].frame_pub.make_sub()

    @staticmethod
    def cam_status_sub(cam_id, blocking=True):
        if blocking:
            while cam_id not in CamCtrl.cv_cams_dict:
                continue
        return CamCtrl.cv_cams_dict[str(cam_id)].status_pub.make_sub()

    @staticmethod
    def handler_cmd_sub(cam_id, blocking=True):
        if blocking:
            while cam_id not in CamCtrl.cv_cam_handlers_dict:
                continue
        return CamCtrl.cv_cam_handlers_dict[str(cam_id)].cmd_pub.make_sub()


import threading

import numpy as np

from .pub_cam import pub_cam_thread
from cvpubsubs.webcam_pub.camctrl import CamCtrl
if False:
    from typing import Union, Tuple, Any, Callable, List

from cvpubsubs.window_sub import SubscriberWindows


def global_cv_display_callback(frame,  # type: np.ndarray
                               cam_id  # type: Union[int, str]
                               ):
    """Default callback for sending frames to the global frame dictionary.

    :param frame: The video or image frame
    :type frame: np.ndarray
    :param cam_id: The video or image source
    :type cam_id: Union[int, str]
    """
    SubscriberWindows.frame_dict[str(cam_id) + "frame"] = frame

display_callbacks = [global_cv_display_callback]

class VideoHandlerThread(threading.Thread):
    "Thread for publishing frames from a video source."

    def __init__(self, video_source=0,  # type: Union[int, str]
                 callbacks=(global_cv_display_callback,),  # type: List[Callable[[np.ndarray, int], Any]]
                 request_size=(-1, -1),  # type: Tuple[int, int]
                 high_speed=True,  # type: bool
                 fps_limit=240  # type: float
                 ):
        """Sets up the main thread loop.

        :param video_source: The video or image source. Integers typically access webcams, while strings access files.
        :type video_source: Union[int, str]
        :param callbacks: A list of operations to be performed on every frame, including publishing.
        :type callbacks: List[Callable[[np.ndarray, int], Any]]
        :param request_size: Requested video size. Actual size may vary, since this is requesting from the hardware.
        :type request_size: Tuple[int, int]
        :param high_speed: If true, use compression to increase speed.
        :type high_speed: bool
        :param fps_limit: Limits frames per second.
        :type fps_limit: float
        """
        super(VideoHandlerThread, self).__init__(target=self.loop, args=())
        if isinstance(video_source, (int, str)):
            self.cam_id = str(video_source)
        elif isinstance(video_source, np.ndarray):
            self.cam_id = str(hash(str(video_source)))
        else:
            raise TypeError(
                "Only strings or ints representing cameras, or numpy arrays representing pictures supported.")
        self.video_source = video_source
        self.callbacks = callbacks
        self.request_size = request_size
        self.high_speed = high_speed
        self.fps_limit = fps_limit

    def loop(self):
        """Continually gets frames from the video publisher, runs callbacks on them, and listens to commands."""
        t = pub_cam_thread(self.video_source, self.request_size, self.high_speed, self.fps_limit)
        while str(self.cam_id) not in CamCtrl.cv_cams_dict:
            continue
        sub_cam = CamCtrl.cam_frame_sub(str(self.cam_id))
        sub_owner = CamCtrl.handler_cmd_sub(str(self.cam_id))
        msg_owner = sub_owner.return_on_no_data = ''
        while msg_owner != 'quit':
            frame = sub_cam.get(blocking=True, timeout=1.0)  # type: np.ndarray
            if frame is not None:
                frame = frame
                for c in self.callbacks:
                    frame = c(frame, self.cam_id)
            msg_owner = sub_owner.get()
        sub_owner.release()
        sub_cam.release()
        CamCtrl.stop_cam(self.cam_id)
        t.join()

    def display(self,
                callbacks=()  # type: List[Callable[[List[np.ndarray]], Any]]
                ):
        """Default display operation. For multiple video sources, please use something outside of this class.

        :param callbacks: List of callbacks to be run on frames before displaying to the screen.
        :type callbacks: List[Callable[[List[np.ndarray]], Any]]
        """
        self.start()
        SubscriberWindows(video_sources=[self.cam_id], callbacks=callbacks).loop()
        self.join()
